#!/usr/bin/env python
#
# Data Quality Checker also known as DQC
#
# Read the README.md for more detailed information
#
#       Barcelona Supercomputing Center - Centro Nacional de Supercomputacion
#       Earth Sciences Department - Computational Earth Sciences
#     
#     Authors: 
#       Joan Sala Calero (joan.sala@bsc.es)
#       Jesus Pena-Izquierdo (jesus.pena@bsc.es)
#       Ivan Jose Cernicharo Ortiz (ivan.cernicharo@bsc.es)

import sys
import logging
import os
import re
import subprocess
import numpy as np
import xarray
import cfgrib
import numpy as np
import pandas as pd
from scipy import stats

# Local imports
from utils.utils_files import *
from utils.utils_xarray import *

"""
Class: Valid ranges checks
---
- Checks data to detect outliers
"""
class valid_ranges_test:

    def __init__(self, logdir, files, varname, vmin, vmax, dt, cz, kstd=10):
        self.testid = 'test06-valid_ranges'
        self.logger = logging.getLogger(self.testid)
        self.logdir = logdir 
        self.files = files
        self.varname = varname     
        self.datatype = dt   
        self.chunksize = cz
        self.stat_columns = ['data_Min', 'data_Max', 'data_Mean', 'data_Std']
        self.hist_columns = [ 'hist_m', 'hist_dev', 'hist_med', 'hist_q1', 'hist_q3', 'hist_IQD', 'hist_skew' ]

        # Number of standard deviations
        self.kstd = kstd
        
        # Automatic versus computed valid ranges
        try:
            self.valid_min = float(vmin)
        except:
            self.valid_min = None
        try:
            self.valid_max = float(vmax)
        except:
            self.valid_max = None

    def adj_boxplot(self,q1,q3,skew, dev, Ks, Ke):
        IQD = q3 - q1
        if IQD == 0:
            IQD = dev
        return q1 - Ke*IQD*np.exp(-Ks*skew1), q3 + Ke*IQD*np.exp(Ks*skew1)

    def std_boxplot(self,m, dev, K):
        return m -dev*K, m + dev*K

    def iqd_boxplot(self,q1,q3,K = 1.5):
        IQD = q3 - q1
        if IQD == 0:
            IQD = dev
        return q1 - K*IQD, q3 + K*IQD

    def hist2array(self,xm, n):
        x = [[ xi]*int(ni) for (xi, ni) in zip(xm, n)]
        return np.hstack(x)

    def hist2array_cut(self,xm, n, rcut = 0.001):
        Nt = np.sum(n)
        ncum = np.cumsum(n/Nt)
        i_nr0i = np.argwhere( ncum <= rcut ).flatten()
        i_nr0d = np.argwhere( ncum >= (1-rcut)).flatten()
        i_nr0 = np.concatenate((i_nr0i, i_nr0d))
        i_nr1 = np.argwhere((ncum > rcut) & (ncum < 1-rcut)).flatten()
        n_r0 = n[i_nr0]
        xm_r0 = xm[i_nr0]
        max_r0id = [n[i_nr0i[-1]], n[i_nr0d[0]]]
        if 0 in max_r0id:
            max_r0id.remove(0)
        k_r0 = np.min(max_r0id)
        print(k_r0)
        n_r1 = np.round(n[i_nr1]/k_r0)
        xm_r1 = xm[i_nr1]
        x_k0 = hist2array(xm_r0, n_r0)
        x_k1 = hist2array(xm_r1, n_r1)
        return x_k0, x_k1    

    """
    Producing statistical indicators of the histogram
    """
    def get_stats_from_hist(self, xm, ni):

        Nt = np.sum(ni)
        ncum = np.cumsum(ni) / Nt
        m = np.sum(xm*ni) / Nt
        dev = (np.sum( ( (xm - m)**2 )*ni ) / Nt )**0.5
        quantiles = [0.25, 0.5, 0.75]
        iquantile = [np.argwhere(ncum > i)[0].item() for i in [0.25, 0.5, 0.75]]
        q1 = xm[iquantile[0]]
        med = xm[iquantile[1]]
        q3 = xm[iquantile[2]]
        IQD = q3 - q1
        skew1 = (m - med) / dev
        
        return [ m, dev, med, q1, q3, IQD, skew1 ]  

    """
    Producing a list of statistical estimators (on the data)
    """
    def get_basic_stats(self, vals):
        
        Min = np.min(vals)
        Max = np.max(vals)
        Mean = np.mean(vals)
        Std = np.std(vals)

        return [ Min, Max, Mean, Std ]

    """
    Loading DATA in memory depending on fileformat | All Timesteps | Filter out NaN
    """
    def load_datafile(self, filepath):

        # Read data into Xarray
        ds = open_dataset(filepath, self.datatype, c_chunks=self.chunksize)        
        arr = get_variable(ds, self.varname)
        
        # Remove NaN values + flatten
        arr = arr[~np.isnan(arr)]
        
        # Data to be a flatten numpy array
        if isinstance(arr, np.ndarray):
            return arr
        else:
            return arr.compute()

    """
    Get Histogram bins initalization. Number of bins + K = number of times (amplification of the histogram bins) 
    """
    def get_stats_files(self):
        # In a CSV file    
        csv_stats = re.sub("([0-9]{8})", "stats.csv", os.path.splitext(self.files[0])[0]) # YYYYMMDD
        csv_hist = re.sub("([0-9]{8})", "hist.csv", os.path.splitext(self.files[0])[0]) # YYYYMMDD
        self.logger.info('Saving stats in {}'.format(os.path.basename(csv_stats)))
        print('INFO: Saving stats in {}'.format(csv_stats))
        self.logger.info('Saving histograms in {}'.format(os.path.basename(csv_hist)))
        print('INFO: Saving histograms in {}'.format(csv_hist))        

        return csv_stats, csv_hist

    """
    Get Histogram bins initalization. Number of bins + K = number of times (amplification of the histogram bins) 
    """
    def get_bins_init(self, f, nbins=250, Kmax=10):
        data = self.load_datafile(f)
        minv = np.min(data)
        maxv = np.max(data)
        dist = abs(maxv-minv)        

        return np.histogram_bin_edges(data, bins=nbins, range=(minv-Kmax*dist, maxv+Kmax*dist))

    """
    Main function
    """
    def check_valid_ranges(self):
       
        # Progress init
        i = 0
        n = len(self.files) # read/calc_stats

        # Store stats information per file       
        bn = [ os.path.basename(f) for f in self.files ] +  ['total']
        df_stats = pd.DataFrame(index=bn, columns=self.stat_columns+self.hist_columns)

        # Init histogram bins
        hist_bins = self.get_bins_init(self.files[0])
        df_hist = pd.DataFrame(index=bn, columns=hist_bins[:-1])
       
        # Generate histograms, store them in a dataframe
        for f in self.files:
            # Progress update
            i+=1
            print_progress(i, n)              
            
            # Read data
            data = self.load_datafile(f)

            # Compute histogram                
            x, b = np.histogram(data, bins=hist_bins)    
            df_hist.loc[os.path.basename(f)] = x

            # Compute stats (of the histogram and the data itself)            
            df_stats.loc[os.path.basename(f)] = self.get_basic_stats(data) + self.get_stats_from_hist(x, b[:-1])
            
            # Free Mem
            del data

        # Compute total histogram        
        df_hist.loc['total'] = df_hist.sum().values

        # Save stats for future use
        csv_stats, csv_hist = self.get_stats_files()
        df_hist.index.name = 'File'
        df_hist.to_csv(csv_hist)
        df_stats.index.name = 'File'
        df_stats.to_csv(csv_stats)

        # Just histogram delivered, unknown ranges for the variable
        res = 'ok'
        msg = ''
        if self.valid_min is None and self.valid_max is None:            
            msg = 'Physical plausible range unknown. Observed range: [ from {} to  {} ]'.format(min(df_stats['data_Min']), min(df_stats['data_Max']))
            self.logger.error(msg)             

        # Valid range estimation (only if defined, well-known variable limits)               
        else:
            for fname, row in df_stats.iterrows():
                # Check ranges          
                if row['data_Min'] < self.valid_min or row['data_Max'] > self.valid_max:
                    self.logger.error('{} contains outliers'.format(fname))
                    self.logger.error('[ {} > {} ] vs [ {} > {} ]'.format(self.valid_min, self.valid_max, row['data_Min'], row['data_Max']))
                    res = 'err'               

            # Conclusion plausible range            
            if res == 'ok':
                msg = 'No outliers found. Plausible range: [ from {} to  {} ]'.format(self.valid_min, self.valid_max)      
            else:
                msg = 'Potential outliers found. Plausible range: [ from {} to  {} ]'.format(self.valid_min, self.valid_max)

        # Final message
        self.logger.info(msg)
        return res, msg

    def main(self):
       
        # Do the actual work
        res, msg = self.check_valid_ranges()

        # Prepare output
        return { 
            'res': res, 
            'msg': msg
        }    
    
if __name__ == '__main__':
    sys.exit()
